@@ -16,6 +16,7 @@ import Language.Haskell.TH.ExpandSyns.SemigroupCompat as Sem
 import Language.Haskell.TH hiding(cxt)
 import qualified Data.Set as Set
 import Data.Generics
+import Data.Maybe
 import Control.Monad
 import Prelude
 
@@ -201,6 +202,14 @@ decIsSyn settings = go
     go (PatSynSigD {}) = no
 #endif
 
+#if MIN_VERSION_template_haskell(2,15,0)
+    go (ImplicitParamBindD {}) = no
+#endif
+
+#if MIN_VERSION_template_haskell(2,16,0)
+    go (KiSigD {}) = no
+#endif
+
     no = return Nothing
 
 #if MIN_VERSION_template_haskell(2,4,0)
@@ -229,11 +238,48 @@ expandSynsWith settings = expandSyns'
       expandSyns' t =
          do
            (acc,t') <- go [] t
-           return (foldl AppT t' acc)
+           return (foldl applyTypeArg t' acc)
+
+#if MIN_VERSION_template_haskell(2,4,0)
+      expandKindSyns' k =
+# if MIN_VERSION_template_haskell(2,8,0)
+         do
+           (acc,k') <- go [] k
+           return (foldl applyTypeArg k' acc)
+# else
+         return k -- No kind variables on old versions of GHC
+# endif
+#endif
+
+      applyTypeArg :: Type -> TypeArg -> Type
+      applyTypeArg f (TANormal x) = f `AppT` x
+      applyTypeArg f (TyArg _x)   =
+#if __GLASGOW_HASKELL__ >= 807
+                                    f `AppKindT` _x
+#else
+                                    -- VKA isn't supported, so
+                                    -- conservatively drop the argument
+                                    f
+#endif
+
+
+      -- Filter the normal type arguments from a list of TypeArgs.
+      filterTANormals :: [TypeArg] -> [Type]
+      filterTANormals = mapMaybe getTANormal
+        where
+          getTANormal :: TypeArg -> Maybe Type
+          getTANormal (TANormal t) = Just t
+          getTANormal (TyArg {})   = Nothing
 
       -- Must only be called on an `x' requiring no expansion
       passThrough acc x = return (acc, x)
 
+      forallAppError :: [TypeArg] -> Type -> Q a
+      forallAppError acc x =
+          fail (packagename++": Unexpected application of the local quantification: "
+                ++show x
+                ++"\n    (to the arguments "++show acc++")")
+
       -- If @go args t = (args', t')@,
       --
       -- Precondition:
@@ -242,7 +288,7 @@ expandSynsWith settings = expandSyns'
       --  All elements of `args'' and `t'' are expanded.
       --  `t' applied to `args' equals `t'' applied to `args'' (up to expansion, of course)
 
-      go :: [Type] -> Type -> Q ([Type], Type)
+      go :: [TypeArg] -> Type -> Q ([TypeArg], Type)
 
       go acc x@ListT = passThrough acc x
       go acc x@ArrowT = passThrough acc x
@@ -254,15 +300,12 @@ expandSynsWith settings = expandSyns'
         t' <- expandSyns' t
         return ([], ForallT ns cxt' t')
 
-      go acc x@(ForallT _ _ _) =
-          fail (packagename++": Unexpected application of the local quantification: "
-                ++show x
-                ++"\n    (to the arguments "++show acc++")")
+      go acc x@ForallT{} = forallAppError acc x
 
       go acc (AppT t1 t2) =
           do
             r <- expandSyns' t2
-            go (r:acc) t1
+            go (TANormal r:acc) t1
 
       go acc x@(ConT n) =
           do
@@ -274,8 +317,8 @@ expandSynsWith settings = expandSyns'
                   then fail (packagename++": expandSynsWith: Underapplied type synonym: "++show(n,acc))
                   else
                       let
-                          substs = zip vars acc
-                          expanded = foldr subst body substs
+                          substs = zip vars (filterTANormals acc)
+                          expanded = doSubsts substs body
                       in
                         go (drop (length vars) acc) expanded
 
@@ -284,11 +327,8 @@ expandSynsWith settings = expandSyns'
       go acc (SigT t kind) =
           do
             (acc',t') <- go acc t
-            return
-              (acc',
-                SigT t' kind
-                -- No expansion needed in kinds (todo: is this correct?)
-              )
+            kind' <- expandKindSyns' kind
+            return (acc', SigT t' kind')
 #endif
 
 #if MIN_VERSION_template_haskell(2,6,0)
@@ -331,6 +371,32 @@ expandSynsWith settings = expandSyns'
       go acc x@(UnboxedSumT _) = passThrough acc x
 #endif
 
+#if MIN_VERSION_template_haskell(2,15,0)
+      go acc (AppKindT t k) =
+          do
+            k' <- expandKindSyns' k
+            go (TyArg k':acc) t
+      go acc (ImplicitParamT n t) =
+          do
+            (acc',t') <- go acc t
+            return (acc',ImplicitParamT n t')
+#endif
+
+#if MIN_VERSION_template_haskell(2,16,0)
+      go [] (ForallVisT ns t) = do
+        t' <- expandSyns' t
+        return ([], ForallVisT ns t')
+
+      go acc x@ForallVisT{} = forallAppError acc x
+#endif
+
+-- | An argument to a type, either a normal type ('TANormal') or a visible
+-- kind application ('TyArg').
+data TypeArg
+  = TANormal Type -- Normal arguments
+  | TyArg    Kind -- Visible kind applications
+  deriving Show
+
 class SubstTypeVariable a where
     -- | Capture-free substitution
     subst :: (Name, Type) -> a -> a
@@ -338,7 +404,7 @@ class SubstTypeVariable a where
 
 
 instance SubstTypeVariable Type where
-  subst (v, t) = go
+  subst vt@(v, t) = go
     where
       go (AppT x y) = AppT (go x) (go y)
       go s@(ConT _) = s
@@ -347,12 +413,13 @@ instance SubstTypeVariable Type where
       go ArrowT = ArrowT
       go ListT = ListT
       go (ForallT vars cxt body) =
-          commonForallCase (v,t) (vars,cxt,body)
+          commonForallCase vt vars $ \vts' vars' ->
+          ForallT vars' (map (doSubsts vts') cxt) (doSubsts vts' body)
 
       go s@(TupleT _) = s
 
 #if MIN_VERSION_template_haskell(2,4,0)
-      go (SigT t1 kind) = SigT (go t1) kind
+      go (SigT t1 kind) = SigT (go t1) (subst vt kind)
 #endif
 
 #if MIN_VERSION_template_haskell(2,6,0)
@@ -384,6 +451,17 @@ instance SubstTypeVariable Type where
       go s@(UnboxedSumT _) = s
 #endif
 
+#if MIN_VERSION_template_haskell(2,15,0)
+      go (AppKindT ty ki) = AppKindT (go ty) (go ki)
+      go (ImplicitParamT n ty) = ImplicitParamT n (go ty)
+#endif
+
+#if MIN_VERSION_template_haskell(2,16,0)
+      go (ForallVisT vars body) =
+          commonForallCase vt vars $ \vts' vars' ->
+          ForallVisT vars' (doSubsts vts' body)
+#endif
+
 -- testCapture :: Type
 -- testCapture =
 --     let
@@ -402,6 +480,10 @@ instance SubstTypeVariable Pred where
     subst s = mapPred (subst s)
 #endif
 
+#if MIN_VERSION_template_haskell(2,4,0) && !MIN_VERSION_template_haskell(2,8,0)
+instance SubstTypeVariable Kind where
+    subst _ = id -- No kind variables on old versions of GHC
+#endif
 
 -- | Make a name (based on the first arg) that's distinct from every name in the second arg
 --
@@ -441,15 +523,16 @@ evades ns t = foldr c [] ns
 --               evade v (AppT (VarT v) (VarT (mkName "fx")))
 
 instance SubstTypeVariable Con where
-  subst (v,t) = go
+  subst vt = go
     where
-      st = subst (v,t)
+      st = subst vt
 
       go (NormalC n ts) = NormalC n [(x, st y) | (x,y) <- ts]
       go (RecC n ts) = RecC n [(x, y, st z) | (x,y,z) <- ts]
       go (InfixC (y1,t1) op (y2,t2)) = InfixC (y1,st t1) op (y2,st t2)
       go (ForallC vars cxt body) =
-          commonForallCase (v,t) (vars,cxt,body)
+          commonForallCase vt vars $ \vts' vars' ->
+          ForallC vars' (map (doSubsts vts') cxt) (doSubsts vts' body)
 #if MIN_VERSION_template_haskell(2,11,0)
       go c@GadtC{} = errGadt c
       go c@RecGadtC{} = errGadt c
@@ -469,16 +552,16 @@ instance HasForallConstruct Con where
 
 
 
-commonForallCase :: (SubstTypeVariable a, HasForallConstruct a) =>
-
-                    (Name,Type)
-                 -> ([TyVarBndr],Cxt,a)
+-- Apply a substitution to something underneath a @forall@. The continuation
+-- argument provides new substitutions and fresh type variable binders to avoid
+-- the outer substitution from capturing the thing underneath the @forall@.
+commonForallCase :: (Name, Type) -> [TyVarBndr]
+                 -> ([(Name, Type)] -> [TyVarBndr] -> a)
                  -> a
-commonForallCase vt@(v,t) (bndrs,cxt,body)
-
+commonForallCase vt@(v,t) bndrs k
             -- If a variable with the same name as the one to be replaced is bound by the forall,
             -- the variable to be replaced is shadowed in the body, so we leave the whole thing alone (no recursion)
-          | v `elem` (tyVarBndrGetName <$> bndrs) = mkForall bndrs cxt body
+          | v `elem` (tyVarBndrGetName <$> bndrs) = k [vt] bndrs
 
           | otherwise =
               let
@@ -487,16 +570,12 @@ commonForallCase vt@(v,t) (bndrs,cxt,body)
                   freshes = evades vars t
                   freshTyVarBndrs = zipWith tyVarBndrSetName freshes bndrs
                   substs = zip vars (VarT <$> freshes)
-                  doSubsts :: SubstTypeVariable b => b -> b
-                  doSubsts x = foldr subst x substs
-
               in
-                mkForall
-                  freshTyVarBndrs
-                  (fmap (subst vt . doSubsts) cxt )
-                  (     (subst vt . doSubsts) body)
-
+                k (vt:substs) freshTyVarBndrs
 
+-- Apply multiple substitutions.
+doSubsts :: SubstTypeVariable a => [(Name, Type)] -> a -> a
+doSubsts substs x = foldr subst x substs
 
 -- | Capture-free substitution
 substInType :: (Name,Type) -> Type -> Type
