@@ -68,6 +68,13 @@ main = do
         [t| Int'' |]
         [t| Int |])
 
+#if MIN_VERSION_template_haskell(2,8,0)
+    putStrLn "Synonyms in kinds"
+    $(mkTest
+        (sigT (conT ''Int) (ConT ''Id `AppT` StarT))
+        (sigT (conT ''Int) StarT))
+#endif
+
     $(do
         reportWarning "No warning about type families should appear after this line." -- TODO: Automate this test with a custom Quasi instance?
         _ <- expandSynsWith noWarnTypeFamilies =<< [t| (DF1 Int', TF1 Int', AT1 Int') |]
